"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
exports.__esModule = true;
exports.catalog = void 0;
var lodash_1 = require("lodash");
var HttpStatusCodes = require("http-status-codes");
var filtering_1 = require("../helpers/filtering");
var ordering_1 = require("../helpers/ordering");
var paging_1 = require("../helpers/paging");
var metadataObjects_1 = require("../helpers/metadataObjects");
var errors = require("../errors/errors");
var ValidationError_1 = require("../ValidationError");
var groups_1 = require("../helpers/groups");
function getItems(groups) {
    return lodash_1.flatMap(groups, function (group) { return __spreadArrays(group.metrics, group.facts, group.attributes); });
}
function filterByTypes(types, items) {
    if (types === void 0) { types = ["metric", "fact", "attribute"]; }
    return items.filter(function (item) { return types.some(function (type) { return item.meta.category === type; }); });
}
function filterByTitleAndTypes(request, items) {
    return lodash_1.flow(lodash_1.partial(filterByTypes, request.catalogRequest.types), lodash_1.partial(filtering_1.filterByTitle, request.catalogRequest.filter, metadataObjects_1.getTitle))(items);
}
function getPipeLine(project, request) {
    return lodash_1.flow(lodash_1.partial(metadataObjects_1.filterByTags, request.catalogRequest), lodash_1.partial(filterByTitleAndTypes, request), lodash_1.partial(filtering_1.filterAvailable, project, request.catalogRequest.bucketItems), lodash_1.partial(ordering_1.orderByTitle, "asc", metadataObjects_1.getTitle));
}
function getBucketItems(request) {
    return lodash_1.get(request, "catalogRequest.bucketItems", []);
}
function getUnavailableCount(project, request, items) {
    var bucketItems = getBucketItems(request);
    if (bucketItems.length === 0) {
        return 0;
    }
    var filteredItems = filterByTitleAndTypes(request, items);
    var availableItems = getPipeLine(project, request)(items);
    return filteredItems.length - availableItems.length;
}
function validate(request) {
    if (Object.keys(request).length !== 1) {
        throw new ValidationError_1.ValidationError(errors.moreThanOneRootElement());
    }
    if (!lodash_1.has(request, "catalogRequest")) {
        throw new ValidationError_1.ValidationError(errors.invalidElementKey("catalogRequest", Object.keys(request)[0]));
    }
    if (Array.isArray(request.catalogRequest)) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestStructureType("ARRAY"));
    }
    if (lodash_1.isString(request.catalogRequest)) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestStructureType(""));
    }
    if (lodash_1.has(request, "catalogRequest.types")) {
        if (!Array.from(request.catalogRequest.types).every(function (type) { return type === "metric" || type === "fact" || type === "attribute"; })) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestTypes());
        }
        if (lodash_1.isString(request.catalogRequest.types)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestTypesStructure(""));
        }
        if (!Array.isArray(request.catalogRequest.types)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestTypesStructure("HASH"));
        }
    }
    if (lodash_1.has(request, "catalogRequest.paging")) {
        if (lodash_1.isString(request.catalogRequest.paging)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestPagingStructure(""));
        }
        if (Array.isArray(request.catalogRequest.paging)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestPagingStructure("ARRAY"));
        }
        if (Object.keys(request.catalogRequest.paging).length !== 2 ||
            (Object.keys(request.catalogRequest.paging).length === 2 &&
                !lodash_1.has(request.catalogRequest.paging, "limit") &&
                !lodash_1.has(request.catalogRequest.paging, "offset"))) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestPaging());
        }
    }
    if (lodash_1.has(request, "catalogRequest.paging.offset") &&
        (lodash_1.isString(request.catalogRequest.paging.offset) ||
            lodash_1.isObject(request.catalogRequest.paging.offset) ||
            Array.isArray(request.catalogRequest.paging.offset))) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestScalarType("paging/offset"));
    }
    if (lodash_1.has(request, "catalogRequest.paging.limit") &&
        (lodash_1.isString(request.catalogRequest.paging.limit) ||
            lodash_1.isObject(request.catalogRequest.paging.limit) ||
            Array.isArray(request.catalogRequest.paging.limit))) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestScalarType("paging/limit"));
    }
    if (lodash_1.has(request, "catalogRequest.filter") &&
        (lodash_1.isObject(request.catalogRequest.filter) || Array.isArray(request.catalogRequest.filter))) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestScalarType("filter"));
    }
    if (lodash_1.has(request, "catalogRequest.requiredDataSets") &&
        (lodash_1.isString(request.catalogRequest.requiredDataSets) ||
            Array.isArray(request.catalogRequest.requiredDataSets))) {
        throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestScalarType("requiredDataSets/DataSetsSelector"));
    }
    if (lodash_1.has(request, "catalogRequest.bucketItems")) {
        if (lodash_1.isString(request.catalogRequest.bucketItems)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestBucketItemsStructure(""));
        }
        if (!Array.isArray(request.catalogRequest.bucketItems)) {
            throw new ValidationError_1.ValidationError(errors.invalidCatalogRequestBucketItemsStructure("HASH"));
        }
    }
}
exports.catalog = {
    register: function (app, project) {
        app.post("/gdc/internal/projects/" + project.project.meta.identifier + "/loadCatalog", function (req, res) {
            var request = req.body;
            try {
                validate(request);
                var groups = groups_1.getGroups(request.catalogRequest.requiredDataSets, project);
                var catalogItems = getPipeLine(project, request)(getItems(groups));
                var _a = lodash_1.get(request, "catalogRequest.paging", {
                    offset: 0,
                    limit: NaN
                }), _b = _a.offset, offset = _b === void 0 ? 0 : _b, limit = _a.limit;
                var sliced = paging_1.getPage(catalogItems, offset, limit);
                var body = {
                    catalogResponse: {
                        catalog: sliced.map(metadataObjects_1.getWrappedMetadataObject),
                        totals: {
                            unavailable: getUnavailableCount(project, request, getItems(groups)),
                            available: catalogItems.length
                        },
                        paging: {
                            offset: offset,
                            count: sliced.length
                        }
                    }
                };
                res.status(HttpStatusCodes.OK).json(body);
            }
            catch (error) {
                if (error instanceof ValidationError_1.ValidationError) {
                    res.status(HttpStatusCodes.BAD_REQUEST).json(error.getValidationError());
                }
                else {
                    res.status(HttpStatusCodes.INTERNAL_SERVER_ERROR).json(error);
                }
            }
        });
        return app;
    }
};
//# sourceMappingURL=catalog.js.map