"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
exports.__esModule = true;
exports.executeAfm = exports.doExecuteAfm = exports.cleanOldCache = exports.hasCachedResult = exports.getFromCache = exports.addToCache = exports.sanitize = exports.getNormalizedExecution = exports.isExecutionValid = exports.isErrorResponse = void 0;
var crypto_1 = require("crypto");
var HttpStatusCodes = require("http-status-codes");
var lodash_1 = require("lodash");
var SchemaAfmExecution_1 = require("../../schema/model/SchemaAfmExecution");
var http_1 = require("../../utils/http");
var log = require("../../utils/log");
var dataResult_1 = require("./executeAfm/dataResult");
var dimensions_1 = require("./executeAfm/dimensions");
var execution_1 = require("./executeAfm/execution");
var getDataSize_1 = require("./executeAfm/getDataSize");
var randomGenerator_1 = require("./executeAfm/helpers/randomGenerator");
var resultSpec_1 = require("./executeAfm/resultSpec");
var sorting_1 = require("./executeAfm/sorting");
var totals_1 = require("./executeAfm/totals");
var measureValueFilters_1 = require("./executeAfm/measureValueFilters");
var rankingFilters_1 = require("./executeAfm/rankingFilters");
//
// Helpers
//
function isErrorResponse(response) {
    return response.error !== undefined;
}
exports.isErrorResponse = isErrorResponse;
function resultSpecContainsMeasureGroup(resultSpec) {
    var dimensions = resultSpec.dimensions;
    if (!dimensions) {
        return false;
    }
    return dimensions.some(function (dimension) { return dimension.itemIdentifiers.includes("measureGroup"); });
}
function isAfmExecutionValid(execution) {
    return isExecutionValid(execution.execution.afm, execution.execution.resultSpec);
}
function isExecutionValid(afm, resultSpec) {
    return !resultSpecContainsMeasureGroup(resultSpec) || (!!afm.measures && afm.measures.length > 0);
}
exports.isExecutionValid = isExecutionValid;
function executionKey(execution) {
    return crypto_1.createHash("md5")
        .update(JSON.stringify(execution))
        .digest("hex");
}
//
// AFM sanitization / normalization
//
function getNormalizedExecution(execution) {
    return {
        execution: {
            afm: execution.execution.afm,
            resultSpec: getNormalizedResultSpec(execution)
        }
    };
}
exports.getNormalizedExecution = getNormalizedExecution;
function getNormalizedResultSpec(execution) {
    var normalizedResultSpec = lodash_1.cloneDeep(execution.execution.resultSpec);
    if (lodash_1.isEmpty(normalizedResultSpec.dimensions)) {
        return __assign(__assign({}, normalizedResultSpec), resultSpec_1.getDefaultResultSpec(execution.execution.afm));
    }
    return normalizedResultSpec;
}
function sanitize(execution) {
    var nativeTotalItems = lodash_1.get(execution, "execution.afm.nativeTotals");
    var nativeTotals = nativeTotalItems ? { nativeTotals: nativeTotalItems } : {};
    return {
        execution: {
            afm: __assign({ measures: lodash_1.get(execution, "execution.afm.measures", []), attributes: lodash_1.get(execution, "execution.afm.attributes", []), filters: lodash_1.get(execution, "execution.afm.filters", []) }, nativeTotals),
            resultSpec: {
                sorts: lodash_1.get(execution, "execution.resultSpec.sorts", []),
                dimensions: lodash_1.get(execution, "execution.resultSpec.dimensions", [])
            }
        }
    };
}
exports.sanitize = sanitize;
/**
 * Store of the generated execution results
 *
 * Keys are represented by MD5 hashes of AFM.IExecution
 */
var cacheResult = new Map();
function addToCache(key, response) {
    var entry = cacheResult.get(key);
    if (entry === undefined) {
        entry = {
            response: response,
            createdAt: Date.now()
        };
        cacheResult.set(key, entry);
    }
    else {
        entry.createdAt = Date.now();
    }
    entry.response = response;
}
exports.addToCache = addToCache;
function getFromCache(key) {
    var result = cacheResult.get(key);
    return result !== undefined ? result.response : undefined;
}
exports.getFromCache = getFromCache;
function hasCachedResult(key) {
    return cacheResult.has(key);
}
exports.hasCachedResult = hasCachedResult;
function deleteFromCache(key) {
    cacheResult["delete"](key);
}
function cleanOldCache(lifetime) {
    if (lifetime === void 0) { lifetime = 240000; }
    var now = Date.now();
    var oldEntries = [];
    cacheResult.forEach(function (entry, key) {
        if (entry.createdAt < now - lifetime) {
            oldEntries.push(key);
        }
    });
    oldEntries.forEach(function (entry) { return deleteFromCache(entry); });
}
exports.cleanOldCache = cleanOldCache;
//
// Result & response creation and generation
//
/**
 * We used to have different generation logic and seed when generating data for stacked charts. While having that
 * logic separate is no longer needed, the code needs to retain backward compatibility in regards to randomly
 * generated values (as they are used in screenshot tests).
 *
 * This function tests whether resultSpec dimensions are such that a non-default random generator (with different
 * seed) should be used.
 *
 * @param execution execution to test dimensions in
 */
function needDifferentSeedForBackwardCompatibility(execution) {
    var dimensions = execution.execution.resultSpec.dimensions;
    return (dimensions.length === 2 &&
        dimensions[0].itemIdentifiers.length === 1 &&
        dimensions[0].itemIdentifiers[0] !== "measureGroup" &&
        dimensions[1].itemIdentifiers.length === 2 &&
        dimensions[1].itemIdentifiers[0] !== "measureGroup" &&
        dimensions[1].itemIdentifiers[1] === "measureGroup");
}
function generateDataResult(project, execution, seed, randomGeneratorPool) {
    var randomGenerator = function (params) {
        return randomGenerator_1.getRandomDataValue(randomGeneratorPool, seed, "other-" + params.localIdentifier);
    };
    if (needDifferentSeedForBackwardCompatibility(execution)) {
        // see comments above for why this is here
        randomGenerator = function () {
            return randomGenerator_1.getRandomDataValue(randomGeneratorPool, seed, "stacked");
        };
    }
    try {
        return dataResult_1.DataResultBuilder.forProject(project)
            .forAfmAndResultSpec(execution)
            .atDate(new Date())
            .withGenerator(randomGenerator)
            .build();
    }
    catch (e) {
        if (e instanceof sorting_1.SortDefinitionError) {
            return {
                error: {
                    statusCode: HttpStatusCodes.BAD_REQUEST,
                    message: e.getMessage()
                }
            };
        }
        throw e;
    }
}
function createExecutionResult(executionResponse, dataResult, totals) {
    var isEmpty = dataResult.isEmpty;
    var total = getDataSize_1.getSizeFromHeaders(dataResult.headerItems);
    var dimensionsCount = total.length;
    var totalsEmpty = lodash_1.flattenDepth(totals, 2).length === 0;
    var totalsNormalized = totalsEmpty ? undefined : totals;
    var headerItems = hasSomeHeaderItems(dataResult.headerItems)
        ? { headerItems: dataResult.headerItems }
        : {};
    return {
        executionResult: {
            executionResult: __assign(__assign({ data: isEmpty ? [] : dataResult.data }, headerItems), { totals: totalsNormalized, paging: {
                    count: total,
                    offset: Array(dimensionsCount).fill(0),
                    total: total
                } })
        },
        executionResponse: executionResponse
    };
}
function hasSomeHeaderItems(array) {
    return lodash_1.flattenDepth(array, Infinity).length > 0;
}
function getExecutionLink(project, execution) {
    var dimensions = lodash_1.get(execution, "execution.resultSpec.dimensions", []);
    var numOfDimensions = !lodash_1.isEmpty(dimensions) ? dimensions.length : "2";
    return ("/gdc/app/projects/" +
        project.project.meta.identifier +
        "/executionResults/" +
        executionKey(execution) +
        "?dimensions=" +
        numOfDimensions);
}
function createExecutionResponse(project, execution, totalTypes) {
    if (totalTypes === void 0) { totalTypes = []; }
    var afm = execution.execution.afm;
    var resultSpec = getNormalizedResultSpec(execution);
    var dimensions = dimensions_1.generateDimensions(project, afm, resultSpec, totalTypes);
    return {
        executionResponse: {
            links: {
                executionResult: getExecutionLink(project, execution)
            },
            dimensions: dimensions
        }
    };
}
function generateExecution(project, execution, seed, randomGeneratorPool) {
    var normalizedExecution = getNormalizedExecution(execution);
    if (!isAfmExecutionValid(normalizedExecution)) {
        return {
            error: {
                statusCode: HttpStatusCodes.BAD_REQUEST,
                message: "Specified 'measureGroup' but no measures found in AFM"
            }
        };
    }
    var dataResult = generateDataResult(project, normalizedExecution, seed, randomGeneratorPool);
    if (isErrorResponse(dataResult)) {
        return dataResult;
    }
    var totalsRandomGenerator = function (params) {
        // Generates deterministic sequence for every type and measure combination. -> Every total cell contains unique
        // value fixed to measure and total type no matter of `measures` bucket items order or totals order.
        return randomGenerator_1.getRandomDataValue(randomGeneratorPool, seed, "totals-" + params.totalType + "-" + params.localIdentifier);
    };
    var totals = totals_1.getTotals(normalizedExecution, dataResult, totalsRandomGenerator);
    if (isErrorResponse(totals)) {
        return totals;
    }
    var executionResponse = createExecutionResponse(project, execution, totals.types);
    return createExecutionResult(executionResponse, dataResult, totals.data);
}
function applyPaging(result, pagingParams) {
    if (result.executionResult.headerItems === undefined) {
        return result;
    }
    var copyResult = lodash_1.cloneDeep(result);
    var start = [pagingParams.offset[0], pagingParams.offset[1]];
    var end = [start[0] + pagingParams.limit[0], start[1] + pagingParams.limit[1]];
    var data = copyResult.executionResult.data;
    var _a = copyResult.executionResult, headerItems = _a.headerItems, paging = _a.paging;
    copyResult.executionResult.headerItems = headerItems.map(function (headers, dim) {
        return headers.map(function (headerItems) {
            return headerItems.slice(start[dim], end[dim]);
        });
    });
    copyResult.executionResult.data = data.slice(start[0], end[0]).map(function (row) {
        if (Array.isArray(row)) {
            return row.slice(start[1], end[1]);
        }
        return row;
    });
    var pagingApplied = copyResult.executionResult.data.length !== result.executionResult.data.length;
    if (pagingApplied) {
        paging.count = getDataSize_1.getSizeFromHeaders(copyResult.executionResult.headerItems);
        paging.offset = pagingParams.offset.slice(0, paging.count.length);
    }
    return copyResult;
}
//
// executeAFM endpoint
//
function logSavedExecution(requestBody, savedAfmExecution) {
    if (savedAfmExecution) {
        log.debug("For request:\n", JSON.stringify(requestBody), "\nsaved execution found:\n", JSON.stringify(savedAfmExecution));
    }
    else {
        log.debug("For request:\n", JSON.stringify(requestBody), "\nNO saved execution found");
    }
}
function getSavedAfmExecution(project, requestBody) {
    var savedAfmExecution = project.afmExecutions.find(function (execution) {
        return execution_1.isRelativelyEqual(requestBody, execution.execution);
    });
    logSavedExecution(requestBody, savedAfmExecution);
    return savedAfmExecution ? savedAfmExecution : null;
}
function toIntArray(param, defaultValue) {
    if (param === undefined || typeof param !== "string") {
        return defaultValue;
    }
    return lodash_1.zip(param.split(",").map(function (str) { return parseInt(str, 10); }), defaultValue).map(function (val) {
        return val[0] === undefined || isNaN(val[0]) ? val[1] : val[0];
    });
}
function extractPagingParameters(request) {
    return {
        limit: toIntArray(request.query.limit, [100, 100]),
        offset: toIntArray(request.query.offset, [0, 0])
    };
}
function getExecutionResultKey(executionResponse) {
    var executionResultLink = executionResponse.executionResponse.links.executionResult;
    var match = executionResultLink.match(/\/executionResults\/(.*)\?/);
    return match[match.length - 1];
}
function doExecuteAfm(project, execution, config) {
    var requestBody = sanitize(execution);
    var savedExecution = getSavedAfmExecution(project, execution);
    if (savedExecution) {
        var executionResult_1 = savedExecution.executionResult, executionResponse_1 = savedExecution.executionResponse;
        var response = executionResponse_1 === undefined
            ? createExecutionResponse(project, requestBody)
            : executionResponse_1;
        if (SchemaAfmExecution_1.isErrorResponse(executionResponse_1)) {
            return {
                error: {
                    statusCode: executionResponse_1,
                    message: "Mocked response Error"
                }
            };
        }
        var measureFilteredExecutionResult = measureValueFilters_1["default"].filterExecutionResult(execution, executionResult_1);
        var rankingFilteredExecutionResult = rankingFilters_1["default"].filterExecutionResult(execution, measureFilteredExecutionResult);
        if (isErrorResponse(rankingFilteredExecutionResult)) {
            return rankingFilteredExecutionResult;
        }
        if (SchemaAfmExecution_1.isExecutionResponseWrapper(response)) {
            addToCache(getExecutionResultKey(response), rankingFilteredExecutionResult);
            return response;
        }
    }
    var randomGeneratorPool = randomGenerator_1.createRandomGeneratorsPool();
    var generatedResult = generateExecution(project, requestBody, config.randomSeed, randomGeneratorPool);
    if (isErrorResponse(generatedResult)) {
        return generatedResult;
    }
    var executionResponse = generatedResult.executionResponse, executionResult = generatedResult.executionResult;
    addToCache(executionKey(requestBody), executionResult);
    return executionResponse;
}
exports.doExecuteAfm = doExecuteAfm;
exports.executeAfm = {
    register: function (app, project, config) {
        app.post("/gdc/app/projects/" + project.project.meta.identifier + "/executeAfm", function (req, res) {
            var result = doExecuteAfm(project, req.body, config);
            if (isErrorResponse(result)) {
                var error = result.error;
                return res.status(error.statusCode).send(error.message);
            }
            return res.status(HttpStatusCodes.CREATED).send(result);
        });
        app.get("/gdc/app/projects/" + project.project.meta.identifier + "/executionResults/:executionId", function (req, res) {
            if (hasCachedResult(req.params.executionId)) {
                var executionResult = getFromCache(req.params.executionId);
                cleanOldCache();
                if (http_1.isHttpStatusCode(executionResult)) {
                    return res.sendStatus(executionResult);
                }
                return res
                    .status(HttpStatusCodes.OK)
                    .send(applyPaging(executionResult, extractPagingParameters(req)));
            }
            return res.sendStatus(HttpStatusCodes.NOT_FOUND);
        });
        return app;
    }
};
//# sourceMappingURL=executeAfm.js.map