"use strict";
exports.__esModule = true;
exports.DataResultBuilder = void 0;
// (C) 2007-2020 GoodData Corporation
var typings_1 = require("@gooddata/typings");
var cartesianProduct_1 = require("../../../utils/cartesianProduct/cartesianProduct");
var executeAfm_1 = require("../executeAfm");
var afm_1 = require("./afm");
var dimensions_1 = require("./dimensions");
var headerItems_1 = require("./headerItems");
var invariant = require("invariant");
var isAttributeHeader = typings_1.Execution.isAttributeHeader;
var isMeasureGroupHeader = typings_1.Execution.isMeasureGroupHeader;
var isMeasureHeaderItem = typings_1.Execution.isMeasureHeaderItem;
var sorting_1 = require("./sorting");
var measureValueFilters_1 = require("./measureValueFilters");
var rankingFilters_1 = require("./rankingFilters");
/**
 * Data result builder is capable of taking any AFM & result spec and generate the headers and data sheet
 * accordingly. This builder can be used for all purposes (simple tabular, stacking, pivot table etc)
 */
var DataResultBuilder = /** @class */ (function () {
    function DataResultBuilder(project) {
        this.date = new Date();
        this.project = project;
    }
    DataResultBuilder.forProject = function (project) {
        return new DataResultBuilder(project);
    };
    DataResultBuilder.prototype.forAfmAndResultSpec = function (execution) {
        var normalizedExecution = executeAfm_1.getNormalizedExecution(execution);
        var afm = normalizedExecution.execution.afm;
        var resultSpec = normalizedExecution.execution.resultSpec;
        this.afm = afm;
        this.resultSpec = resultSpec;
        this.execution = {
            execution: {
                afm: afm,
                resultSpec: resultSpec
            }
        };
        return this;
    };
    DataResultBuilder.prototype.withGenerator = function (generator) {
        this.generator = generator;
        return this;
    };
    DataResultBuilder.prototype.atDate = function (date) {
        this.date = date;
        return this;
    };
    DataResultBuilder.prototype.build = function () {
        this.checkInputs();
        var dims = dimensions_1.generateDimensions(this.project, this.afm, this.resultSpec);
        var measureHeaderItems = [
            this.getExpandedMeasuresFromRows(dims),
            this.getExpandedMeasuresFromCols(dims),
        ];
        var resultHeaderItems = this.generateHeaders(dims);
        var resultData = this.generateData(resultHeaderItems, measureHeaderItems);
        var dataResult = { headerItems: resultHeaderItems, data: resultData };
        var measureValueFilteredDataResult = measureValueFilters_1["default"].filterDataResult(this.execution, dataResult);
        var rankedFilteredDataResult = rankingFilters_1["default"].filterDataResult(this.execution, measureValueFilteredDataResult);
        if (executeAfm_1.isErrorResponse(rankedFilteredDataResult)) {
            return rankedFilteredDataResult;
        }
        var data = rankedFilteredDataResult.data, headerItems = rankedFilteredDataResult.headerItems;
        var result = {
            dims: dims,
            headerItems: headerItems,
            measureHeaderItems: measureHeaderItems,
            data: data,
            isEmpty: !data || !data.length || (data.length > 0 && data[0].length === 0)
        };
        // sort in place
        this.sort(result);
        return result;
    };
    DataResultBuilder.prototype.checkInputs = function () {
        invariant(this.afm !== undefined, "AFM must be defined");
        invariant(this.resultSpec !== undefined, "ResultSpec must be defined");
        invariant(this.generator !== undefined, "Random value generator must be defined");
    };
    /**
     * Generates result headers for each dimension (total of 2).
     *
     * 1. Each dimension contains a list of header items;
     * 2. For each header item in the list:
     *    - if attribute expand to attribute elements
     *    - if measure group expand to all measures in the group
     * 3. This produces a array X = [Y1, Y2, ... Yn] where each Yn is array and N = number of headers in the dim
     * 4. Do a cartesian between elementf of X
     * 5. This produces an array A = [B1, B2, ... Bm] where each Bm is a resulting combination produced by cartesian.
     *    Here M = * (Yn.len) and cardinality of each Bm is N
     * 6. Finally do rotation on A, leading to R = [H1, H2, ... Hn] where N is number of headers in the dim and
     *    cardinality of each Hn is M
     *
     * The resulting list R is the format ready to be returned as a result.
     *
     * @param dims result dimensions
     */
    DataResultBuilder.prototype.generateHeaders = function (dims) {
        var _this = this;
        return dims.map(function (dim) {
            var uniqueHeaderItems = dim.headers.map(function (header) {
                if (isAttributeHeader(header)) {
                    var df = afm_1.getDisplayFormByQualifier(_this.project, {
                        identifier: header.attributeHeader.identifier
                    });
                    return headerItems_1.getAttributeHeaderItemsByDisplayForm(_this.project, _this.afm.filters, df, _this.date);
                }
                else if (isMeasureGroupHeader(header)) {
                    return header.measureGroupHeader.items.map(function (item, idx) {
                        return {
                            measureHeaderItem: {
                                name: item.measureHeaderItem.name,
                                order: idx
                            }
                        };
                    });
                }
            });
            return cartesianProduct_1.rotatedCartesianProduct(uniqueHeaderItems);
        });
    };
    /**
     * Generates an expanded array of measure header items for particular dimension. The cardinality of the
     * resulting array is same as cardinality of header items in the dimension that contains measureGroup.
     *
     * The algorithm here is similar to generateHeaders, the only difference in the return value:
     *
     * - Measure group is NOT expanded into IResultMeasureHeaderItem but into IMeasureHeaderItem - this structure
     *   contains all necessary detail to generate values for the measure (as opposed to result item which contains
     *   name and order)
     *
     * - The result is just the array of IMeasureHeaderItem = one particular subarray
     *
     * @param dim result dimensions
     */
    DataResultBuilder.prototype.getExpandedMeasureItems = function (dim) {
        var _this = this;
        var expandedHeaderItems = cartesianProduct_1.rotatedCartesianProduct(dim.headers.map(function (header) {
            if (isAttributeHeader(header)) {
                var df = afm_1.getDisplayFormByQualifier(_this.project, {
                    identifier: header.attributeHeader.identifier
                });
                return headerItems_1.getAttributeHeaderItemsByDisplayForm(_this.project, _this.afm.filters, df, _this.date);
            }
            else if (isMeasureGroupHeader(header)) {
                return header.measureGroupHeader.items;
            }
        }));
        return expandedHeaderItems.find(function (headers) {
            return headers.length > 0 && isMeasureHeaderItem(headers[0]);
        });
    };
    DataResultBuilder.prototype.getExpandedMeasuresFromRows = function (dims) {
        var measureGroup = dims[0].headers.find(function (header) {
            return isMeasureGroupHeader(header);
        });
        if (measureGroup === undefined) {
            return;
        }
        return this.getExpandedMeasureItems(dims[0]);
    };
    DataResultBuilder.prototype.getExpandedMeasuresFromCols = function (dims) {
        if (dims.length < 2) {
            return [];
        }
        var measureGroup = dims[1].headers.find(function (header) {
            return isMeasureGroupHeader(header);
        });
        if (measureGroup === undefined) {
            return;
        }
        return this.getExpandedMeasureItems(dims[1]);
    };
    DataResultBuilder.prototype.generateRowData = function (rowMeasures, columnHeaders) {
        var _this = this;
        /*
         * Measures are in the row dimension - one measure per row. Iterate rows (= measures) and generate
         * random measure value into each column.
         */
        return rowMeasures.map(function (measure, row) {
            if (columnHeaders === undefined) {
                /*
                 * One-dimensional result requested - return array with measure values
                 */
                return _this.generator({
                    row: row,
                    column: 0,
                    localIdentifier: measure.measureHeaderItem.localIdentifier
                });
            }
            else if (columnHeaders.length === 0) {
                /*
                 * Two-dimension result requested - but second dimension is empty
                 */
                return [
                    _this.generator({
                        row: row,
                        column: 0,
                        localIdentifier: measure.measureHeaderItem.localIdentifier
                    }),
                ];
            }
            return columnHeaders[0].map(function (_, column) {
                return _this.generator({
                    row: row,
                    column: column,
                    localIdentifier: measure.measureHeaderItem.localIdentifier
                });
            });
        });
    };
    DataResultBuilder.prototype.generateColumnData = function (columnMeasures, rowHeaders) {
        var _this = this;
        if (rowHeaders.length === 0) {
            /*
             * Row dimension is empty - generate a single row with all measure values
             */
            return [
                columnMeasures.map(function (measure, column) {
                    return _this.generator({
                        row: 0,
                        column: column,
                        localIdentifier: measure.measureHeaderItem.localIdentifier
                    });
                }),
            ];
        }
        /*
         * Measures are in the column dimension - one measure per column. Iterate rows (= attributes) and
         * generate random value for the measure in that column.
         */
        return rowHeaders[0].map(function (_, row) {
            return columnMeasures.map(function (measure, column) {
                return _this.generator({
                    row: row,
                    column: column,
                    localIdentifier: measure.measureHeaderItem.localIdentifier
                });
            });
        });
    };
    DataResultBuilder.prototype.generateData = function (headers, measureHeaders) {
        var rowMeasures = measureHeaders[0], columnMeasures = measureHeaders[1];
        if (rowMeasures !== undefined) {
            // ... then generate row data => values for one measure are in all cols in the row
            return this.generateRowData(rowMeasures, headers[1]);
        }
        if (columnMeasures !== undefined) {
            // ... then generate column data => values for measure are in particular col(s) in all rows
            return this.generateColumnData(columnMeasures, headers[0]);
        }
        // having no measures is valid too, in that case return empty data
        return [];
    };
    DataResultBuilder.prototype.sort = function (result) {
        new sorting_1.FunSort(result).run(this.resultSpec.sorts);
    };
    return DataResultBuilder;
}());
exports.DataResultBuilder = DataResultBuilder;
//# sourceMappingURL=dataResult.js.map