"use strict";
exports.__esModule = true;
exports.getAttributeHeaderItemsByDisplayForm = exports.getFilteredAttributeElements = void 0;
// (C) 2007-2019 GoodData Corporation
var typings_1 = require("@gooddata/typings");
var date_fns_1 = require("date-fns");
var lodash_1 = require("lodash");
var MockProject_1 = require("../../../model/MockProject");
var routes_1 = require("../../../route/routes");
var ElementsBuilder_1 = require("../../../schema/builders/ElementsBuilder");
var DateRange_1 = require("../../../utils/model/DateRange");
var afm_1 = require("./afm");
function getAttributeElementsReducerDef(filter) {
    if (typings_1.ExecuteAFM.isPositiveAttributeFilter(filter)) {
        return {
            displayForm: filter.positiveAttributeFilter.displayForm,
            inverted: false,
            attributeElements: filter.positiveAttributeFilter["in"]
        };
    }
    else if (typings_1.ExecuteAFM.isNegativeAttributeFilter(filter)) {
        return {
            displayForm: filter.negativeAttributeFilter.displayForm,
            inverted: true,
            attributeElements: filter.negativeAttributeFilter.notIn
        };
    }
    return;
}
function AttributeElementsPredicate(attributeElements) {
    if (typings_1.ExecuteAFM.isAttributeElementsArray(attributeElements)) {
        return function (element) {
            return attributeElements.indexOf(element.element.uri) > -1;
        };
    }
    else if (typings_1.ExecuteAFM.isAttributeElementsByRef(attributeElements)) {
        return function (element) {
            return attributeElements.uris.indexOf(element.element.uri) > -1;
        };
    }
    else if (typings_1.ExecuteAFM.isAttributeElementsByValue(attributeElements)) {
        return function (element) {
            return attributeElements.values.indexOf(element.element.title) > -1;
        };
    }
    return function (_) { return false; };
}
/**
 * Given AFM filter, creates reducer for ElementMap which will filter out elements of the filtered attribute.
 * If given AFM filter is not an attribute element filter, then the resulting reducer will be noop.
 *
 * @param filter AFM filter, any type of filter
 * @param attributeResolver function which resolved attribute display form to attribute
 * @constructor
 */
function AttributeElementsReducer(filter, attributeResolver) {
    var def = getAttributeElementsReducerDef(filter);
    if (def === undefined) {
        // doesn't look like attr element filter -> no element filtering
        return function (elements) { return elements; };
    }
    var attr = attributeResolver(def.displayForm);
    var predicate = AttributeElementsPredicate(def.attributeElements);
    return function (elementsMap) {
        var elements = elementsMap.get(attr.meta.identifier).filter(function (element) {
            return !def.inverted ? predicate(element) : !predicate(element);
        });
        return elementsMap.set(attr.meta.identifier, elements);
    };
}
/**
 * Applies attribute filters.
 *
 * Note: exposed for testability
 *
 * @param project mock project with element infos
 * @param filters filters to apply
 *
 */
function getFilteredAttributeElements(project, filters) {
    var attributeResolver = function (displayFormQualifier) {
        var displayForm = afm_1.getDisplayFormByQualifier(project, displayFormQualifier);
        return MockProject_1.getAttributeByUri(project, displayForm.content.formOf);
    };
    return filters
        .map(function (filter) { return AttributeElementsReducer(filter, attributeResolver); })
        .reduce(function (filteredElements, reducer) { return reducer(filteredElements); }, lodash_1.cloneDeep(project.elements));
}
exports.getFilteredAttributeElements = getFilteredAttributeElements;
function addToDate(date, granularity, value) {
    switch (granularity) {
        case "GDC.time.date":
            return date_fns_1.addDays(date, value);
        case "GDC.time.week_us":
            return date_fns_1.addWeeks(date, value);
        case "GDC.time.month":
            return date_fns_1.addMonths(date, value);
        case "GDC.time.quarter":
            return date_fns_1.addQuarters(date, value);
        case "GDC.time.year":
            return date_fns_1.addYears(date, value);
    }
}
function getStartOfGranularity(date, granularity) {
    switch (granularity) {
        case "GDC.time.date":
            return date_fns_1.startOfDay(date);
        case "GDC.time.week_us":
            return date_fns_1.startOfWeek(date);
        case "GDC.time.month":
            return date_fns_1.startOfMonth(date);
        case "GDC.time.quarter":
            return date_fns_1.startOfQuarter(date);
        case "GDC.time.year":
            return date_fns_1.startOfYear(date);
    }
}
function isCurrentUnit(from, to) {
    return from === 0 && to === 0;
}
function dateRangeFromRelativeFilter(filter, currentDate) {
    var _a = filter.relativeDateFilter, from = _a.from, to = _a.to, granularity = _a.granularity;
    if (isCurrentUnit(from, to)) {
        var startDate_1 = getStartOfGranularity(currentDate, granularity);
        var finalDate_1 = currentDate;
        return DateRange_1.DateRange.fromDates(startDate_1, finalDate_1);
    }
    var startDate = addToDate(currentDate, granularity, from);
    var finalDate = addToDate(currentDate, granularity, to);
    return DateRange_1.DateRange.fromDates(startDate, finalDate);
}
function getAttributeHeaderItemsByDateDisplayForm(project, filters, displayForm, date) {
    // Get date attribute by display form
    var dateAttribute = MockProject_1.getDateAttributes(project).find(function (dateAttribute) { return dateAttribute.displayForms[0].meta.identifier === displayForm.meta.identifier; });
    return getAttributeHeaderItemsByDateAttribute(project, filters, dateAttribute, date);
}
function getAttributeHeaderItemsByDateAttribute(project, filters, dateAttribute, date) {
    // Find date dataset by attribute id
    var dateDataSet = MockProject_1.getDateDataSets(project).find(function (dateDataSet) {
        return dateDataSet.availableDateAttributes.some(function (attribute) {
            return attribute.meta.identifier === dateAttribute.meta.identifier;
        });
    });
    var dateDataSetDateRange = project.dateRanges.get(dateDataSet.meta.identifier);
    var filteredDateRange = filters.reduce(function (tmpRange, filter) {
        if (typings_1.ExecuteAFM.isAbsoluteDateFilter(filter)) {
            var _a = filter.absoluteDateFilter, from = _a.from, to = _a.to;
            return tmpRange.intersect(new DateRange_1.DateRange(from, to));
        }
        if (typings_1.ExecuteAFM.isRelativeDateFilter(filter)) {
            var dateRange = dateRangeFromRelativeFilter(filter, date);
            return tmpRange.intersect(dateRange);
        }
        return tmpRange;
    }, dateDataSetDateRange);
    var dateAttributeElementValues = ElementsBuilder_1.generateDateAttributeElementValues(dateAttribute.type, filteredDateRange);
    var dateAttributeElements = dateAttributeElementValues.map(function (value, i) { return ({
        element: {
            title: value,
            uri: routes_1.elementUri(project.project.meta.identifier, dateAttribute.meta.identifier, "" + (i + 1))
        }
    }); });
    return dateAttributeElements.map(elementToAttributeHeaderItem);
}
function getAttributeHeaderItemsByAttributeDisplayForm(project, filters, displayForm) {
    var attributesMap = getFilteredAttributeElements(project, filters);
    var attribute = MockProject_1.getAttributeByUri(project, displayForm.content.formOf);
    var attributeElements = attributesMap.get(attribute.meta.identifier);
    return attributeElements.map(elementToAttributeHeaderItem);
}
function getAttributeHeaderItemsByDisplayForm(project, filters, displayForm, date) {
    if (MockProject_1.isDateDisplayForm(project, displayForm)) {
        return getAttributeHeaderItemsByDateDisplayForm(project, filters, displayForm, date);
    }
    return getAttributeHeaderItemsByAttributeDisplayForm(project, filters, displayForm);
}
exports.getAttributeHeaderItemsByDisplayForm = getAttributeHeaderItemsByDisplayForm;
function elementToAttributeHeaderItem(_a) {
    var element = _a.element;
    return {
        attributeHeaderItem: {
            uri: element.uri,
            name: element.title
        }
    };
}
//# sourceMappingURL=headerItems.js.map