"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
exports.__esModule = true;
// (C) 2019-2020 GoodData Corporation
var typings_1 = require("@gooddata/typings");
var lodash_1 = require("lodash");
var execution_1 = require("./execution");
var executeAfm_1 = require("../executeAfm");
var getDataSize_1 = require("./getDataSize");
var isMeasureHeaderItem = typings_1.Execution.isMeasureHeaderItem;
var isMeasureValueFilter = typings_1.ExecuteAFM.isMeasureValueFilter;
var REMOVED_VALUE_PLACEHOLDER = "#REMOVED_VALUE#";
function findLocalMeasureValueFilters(filters) {
    if (lodash_1.isEmpty(filters)) {
        return [];
    }
    return filters.filter(function (filter) {
        return isMeasureValueFilter(filter) && execution_1.isILocalIdentifierQualifier(filter.measureValueFilter.measure);
    });
}
function findIndexOfMeasuresDimension(headerItems) {
    return headerItems.findIndex(function (dimension) {
        return dimension.some(function (headers) {
            return headers.some(function (header) { return isMeasureHeaderItem(header); });
        });
    });
}
function findMeasureResultHeaders(measureDimension) {
    return measureDimension.find(function (headers) { return headers.length > 0 && isMeasureHeaderItem(headers[0]); });
}
function findReferencedMeasureAfmIndex(filter, measures) {
    var measureQualifier = filter.measureValueFilter.measure;
    return measures.findIndex(function (measure) { return measure.localIdentifier === measureQualifier.localIdentifier; });
}
/**
 * Apply measure value filters from provided execution to provided dataResult generated for the execution.
 * The function filters the row or column measure data and removes attribute headers that no longer
 * have values.
 *
 * @param execution
 * @param dataResult
 * @return filtered execution result if matching filters were found in execution
 */
function filterDataResult(execution, dataResult) {
    var afm = execution.execution.afm;
    var measures = afm.measures, filters = afm.filters;
    if (lodash_1.isEmpty(measures) || lodash_1.isEmpty(filters)) {
        return dataResult;
    }
    var measureValueFilters = findLocalMeasureValueFilters(filters);
    if (measureValueFilters.length === 0) {
        return dataResult;
    }
    var errorResponse = validateExecution(execution);
    if (errorResponse) {
        return errorResponse;
    }
    var headerItems = dataResult.headerItems, data = dataResult.data;
    var measureDimensionIndex = findIndexOfMeasuresDimension(headerItems);
    var measuresAreInRow = measureDimensionIndex === 0;
    var attributeDimensionIndex = measuresAreInRow ? 1 : 0;
    var measureHeaders = findMeasureResultHeaders(headerItems[measureDimensionIndex]);
    var getRemovedValuesIfSomeValueIsRemoved = function (data) {
        return data.includes(REMOVED_VALUE_PLACEHOLDER)
            ? new Array(data.length).fill(REMOVED_VALUE_PLACEHOLDER)
            : data;
    };
    var filterDataRow = function (dataRow, filter, filteredMeasureIndex) {
        // split values into groups per repeated measures ([m1, m2, m1, m2] => [[m1, m2], [m1, m2]])
        var filteredMeasureGroups = lodash_1.chunk(dataRow, measures.length).map(function (measureGroupValues) {
            // replace values that does not match filter with removed value placeholder in every group
            var processedGroup = measureGroupValues.map(function (value, columnIndex) {
                var currentMeasureIndex = measureHeaders[columnIndex].measureHeaderItem.order;
                return currentMeasureIndex === filteredMeasureIndex
                    ? getMatchingValueOrRemovedValuePlaceholder(value, filter)
                    : value;
            });
            // replace all values in group with removed value placeholder if at least one value is removed value placeholder (was filtered out)
            return getRemovedValuesIfSomeValueIsRemoved(processedGroup);
        });
        // put processed groups to the original shape ([[m1, m2], [m1, m2]] => [m1, m2, m1, m2])
        return lodash_1.flattenDeep(filteredMeasureGroups);
    };
    // rotate the data to have measures in columns and attributes in rows if it is not the case
    var columnData = measuresAreInRow ? rotateMatrix(data) : data;
    // replace values not matching the filters with removed value placeholder
    var validDataResultValues = measureValueFilters.reduce(function (dataResult, filter) {
        var measureIndex = findReferencedMeasureAfmIndex(filter, measures);
        return dataResult.map(function (dataRow) { return filterDataRow(dataRow, filter, measureIndex); });
    }, columnData);
    // remove row attribute headers that have every data value set as removed value placeholder in the row
    var keepRow = function (_, rowIndex) {
        return !validDataResultValues[rowIndex].every(function (columnValue) { return columnValue === REMOVED_VALUE_PLACEHOLDER; });
    };
    var filteredAttributesHeaders = headerItems[attributeDimensionIndex].map(function (rowAttributes) {
        return rowAttributes.filter(keepRow);
    });
    // remove column attribute headers that have every data value set as removed value placeholder in the column
    var keepColumn = function (_, columnIndex) {
        return !validDataResultValues.every(function (dataRow) { return dataRow[columnIndex] === REMOVED_VALUE_PLACEHOLDER; });
    };
    var filteredMeasureHeaders = headerItems[measureDimensionIndex].map(function (columnAttributes) {
        return columnAttributes.filter(keepColumn);
    });
    // remove data rows and columns that contain only removed value placeholder
    var filteredDataRows = validDataResultValues
        .filter(keepRow)
        .map(function (dataRow) { return dataRow.filter(keepColumn); });
    var replaceRemovedValuesByNull = function (value) {
        return value === REMOVED_VALUE_PLACEHOLDER ? null : value;
    };
    // replace removed value placeholder by null values
    var filteredDataRowsWithNullValues = filteredDataRows.map(function (dataRow) {
        return dataRow.map(replaceRemovedValuesByNull);
    });
    // rebuild header item dimensions
    var filteredHeaderItems = [];
    filteredHeaderItems[attributeDimensionIndex] = filteredAttributesHeaders;
    filteredHeaderItems[measureDimensionIndex] = filteredMeasureHeaders;
    return {
        headerItems: filteredHeaderItems,
        data: measuresAreInRow
            ? rotateMatrix(filteredDataRowsWithNullValues) // rotate data back to its original shape
            : filteredDataRowsWithNullValues
    };
}
function rotateMatrix(matrix) {
    return matrix.length === 0 ? matrix : matrix[0].map(function (_, index) { return matrix.map(function (row) { return row[index]; }); });
}
function isComparisonCondition(condition) {
    return !lodash_1.isEmpty(condition) && condition.comparison !== undefined;
}
function isRangeCondition(condition) {
    return !lodash_1.isEmpty(condition) && condition.range !== undefined;
}
function getMatchingValueOrRemovedValuePlaceholder(afmResultValue, filter) {
    return isMatchingFilter(afmResultValue, filter) ? afmResultValue : REMOVED_VALUE_PLACEHOLDER;
}
function transformNullValues(testedValue, treatNullValuesAs) {
    if (treatNullValuesAs === undefined || testedValue !== null) {
        return testedValue;
    }
    return treatNullValuesAs;
}
function isMatchingFilter(afmResultValue, filter) {
    var condition = filter.measureValueFilter.condition;
    var testedValue = afmResultValue === null ? null : Number(afmResultValue);
    if (isComparisonCondition(condition)) {
        var _a = condition.comparison, operator = _a.operator, value = _a.value, treatNullValuesAs = _a.treatNullValuesAs;
        testedValue = transformNullValues(testedValue, treatNullValuesAs);
        if (testedValue === null) {
            return false;
        }
        switch (operator) {
            case "GREATER_THAN":
                return testedValue > value;
            case "GREATER_THAN_OR_EQUAL_TO":
                return testedValue >= value;
            case "LESS_THAN":
                return testedValue < value;
            case "LESS_THAN_OR_EQUAL_TO":
                return testedValue <= value;
            case "EQUAL_TO":
                return testedValue === value;
            case "NOT_EQUAL_TO":
                return testedValue !== value;
            default:
                return true;
        }
    }
    if (isRangeCondition(condition)) {
        var _b = condition.range, operator = _b.operator, from = _b.from, to = _b.to, treatNullValuesAs = _b.treatNullValuesAs;
        testedValue = transformNullValues(testedValue, treatNullValuesAs);
        if (testedValue === null) {
            return false;
        }
        switch (operator) {
            case "BETWEEN":
                return from <= testedValue && testedValue <= to;
            case "NOT_BETWEEN":
                return testedValue < from || to < testedValue;
            default:
                return true;
        }
    }
    return true;
}
function buildBadRequestResponse(message) {
    return {
        error: {
            statusCode: 400,
            message: message
        }
    };
}
function eachMeasureValueFilterReferencesValidLocalMeasure(measures, filters) {
    return filters.every(function (filter) { return findReferencedMeasureAfmIndex(filter, measures) >= 0; });
}
function eachMeasureHasMaximumOneMeasureValueFilter(filters) {
    var referencedMeasureIdentifiers = filters.map(function (filter) { return filter.measureValueFilter.measure.localIdentifier; });
    return referencedMeasureIdentifiers.length === lodash_1.uniq(referencedMeasureIdentifiers).length;
}
function validateExecution(execution) {
    var afm = execution.execution.afm;
    if (lodash_1.isEmpty(afm.attributes)) {
        return buildBadRequestResponse("The insight is filtered by a measure value but it is not sliced by any attribute!");
    }
    var measureValueFilters = findLocalMeasureValueFilters(afm.filters);
    if (!eachMeasureValueFilterReferencesValidLocalMeasure(afm.measures, measureValueFilters)) {
        return buildBadRequestResponse("Some measure value filters reference a measure that was not found in AFM!");
    }
    if (!eachMeasureHasMaximumOneMeasureValueFilter(measureValueFilters)) {
        return buildBadRequestResponse("Some measure is referenced by multiple measure value filters!");
    }
}
function isExecutionResult(result) {
    return !lodash_1.isEmpty(result) && result.executionResult !== undefined;
}
function buildDataResult(result) {
    var _a = result.executionResult, headerItems = _a.headerItems, data = _a.data;
    return {
        headerItems: headerItems,
        data: data
    };
}
function buildExecutionResult(originalResponse, dataResult) {
    var executionResult = originalResponse.executionResult;
    var headerItems = dataResult.headerItems, data = dataResult.data;
    var total = getDataSize_1.getSizeFromHeaders(headerItems);
    return {
        executionResult: __assign(__assign({}, executionResult), { headerItems: headerItems,
            data: data, paging: __assign(__assign({}, executionResult.paging), { count: total, total: total }) })
    };
}
/**
 * Apply measure value filters from provided execution to provided execution result stored for execution.
 * The function filters the row or column measure data and removes attribute headers that no longer
 * have values. It also updates the paging information.
 *
 * @param execution
 * @param response
 * @return filtered execution result if matching filters were found in execution
 */
function filterExecutionResult(execution, response) {
    if (executeAfm_1.isErrorResponse(response) || !isExecutionResult(response)) {
        return response;
    }
    var dataResult = filterDataResult(execution, buildDataResult(response));
    if (executeAfm_1.isErrorResponse(dataResult)) {
        return dataResult;
    }
    return lodash_1.isEqual(dataResult.data, response.executionResult.data)
        ? response
        : buildExecutionResult(response, dataResult);
}
exports["default"] = {
    filterDataResult: filterDataResult,
    filterExecutionResult: filterExecutionResult
};
//# sourceMappingURL=measureValueFilters.js.map