"use strict";
exports.__esModule = true;
exports.createComparator = exports.FunSort = exports.SortDefinitionError = void 0;
// (C) 2007-2019 GoodData Corporation
var typings_1 = require("@gooddata/typings");
var lodash_1 = require("lodash");
var isAttributeSortItem = typings_1.ExecuteAFM.isAttributeSortItem;
var isMeasureLocatorItem = typings_1.ExecuteAFM.isMeasureLocatorItem;
var isMeasureSortItem = typings_1.ExecuteAFM.isMeasureSortItem;
var isAttributeHeader = typings_1.Execution.isAttributeHeader;
var isAttributeHeaderItem = typings_1.Execution.isAttributeHeaderItem;
var ROW_DIM = 0;
var COL_DIM = 1;
/**
 * This is thrown in case sort locators are not valid.
 */
var SortDefinitionError = /** @class */ (function () {
    function SortDefinitionError(message) {
        this.message = message;
        this.message = message;
    }
    SortDefinitionError.prototype.getMessage = function () {
        return this.message;
    };
    return SortDefinitionError;
}());
exports.SortDefinitionError = SortDefinitionError;
/**
 * Fun sort applies sortItems from resultSpec to headers and data. The sorting works as follows:
 *
 *   1. Create sort actions = definitions of what to sort on; these are derived from sort items and reflect
 *      multi-dimensional nature of sort. The actions specify index of row or column to sort on.
 *
 *   2. Sort actions are performed by dimension - row sorting is done first, then column sorting. The sort is done
 *      in multiple steps:
 *
 *      1.  Extract sort entries = values from result to sort by + original index of the values
 *      2.  Sort the entries
 *      4.  Obtain sort vectors = current entry index => original index in the entry
 *      5.  Apply sort vectors on both data and headers
 *
 * This sorting code seems to be able to deal with any combination that the real backend can accept.
 */
var FunSort = /** @class */ (function () {
    function FunSort(result) {
        this.result = result;
        this.result = result;
    }
    FunSort.prototype.run = function (sortItems) {
        if (sortItems === undefined || sortItems.length === 0) {
            return;
        }
        var sortActions = this.createSortActions(sortItems);
        // Calculate sort vectors first => no chance indexes in sort actions will get hosed
        var rowVectors = this.sortVector(sortActions, ROW_DIM);
        var columnVectors = this.sortVector(sortActions, COL_DIM);
        // Finally do the sort
        this.sort([rowVectors, columnVectors]);
    };
    FunSort.prototype.createSortActions = function (sortItems) {
        var _this = this;
        var sortActions = [[], []];
        var _a = this.result, dims = _a.dims, measureHeaderItems = _a.measureHeaderItems;
        sortItems.forEach(function (item) {
            if (isAttributeSortItem(item)) {
                /*
                 * Attribute sorts are always done on the header items (since that's where attr elements are). The
                 * code needs to figure out on which dimension is the sort happening - to achieve this code looks
                 * at result dimensions
                 */
                var match = dims.map(function (dim) {
                    return dim.headers.findIndex(function (header) {
                        return (isAttributeHeader(header) &&
                            header.attributeHeader.localIdentifier ===
                                item.attributeSortItem.attributeIdentifier);
                    });
                });
                if (match[ROW_DIM] > -1) {
                    sortActions[ROW_DIM].push({
                        sortIndex: findHeaderGroupIndexForItem(item, dims[ROW_DIM]),
                        direction: item.attributeSortItem.direction,
                        type: "headers"
                    });
                }
                else if (match[COL_DIM] > -1) {
                    sortActions[COL_DIM].push({
                        sortIndex: findHeaderGroupIndexForItem(item, dims[1]),
                        direction: item.attributeSortItem.direction,
                        type: "headers"
                    });
                }
                else {
                    throw new SortDefinitionError("Bad sort - attribute " + item.attributeSortItem.attributeIdentifier + " not in AFM");
                }
            }
            else if (isMeasureSortItem(item)) {
                /*
                 * Measure sorts are done on the data itself. The code needs to figure out which dimension to
                 * sort on.
                 */
                if (measureHeaderItems[ROW_DIM] !== undefined) {
                    var sortIndex = _this.findSortIndex(item, ROW_DIM);
                    sortActions[COL_DIM].push({
                        sortIndex: sortIndex,
                        direction: item.measureSortItem.direction,
                        type: "data"
                    });
                }
                else if (measureHeaderItems[COL_DIM] !== undefined) {
                    var sortIndex = _this.findSortIndex(item, COL_DIM);
                    sortActions[ROW_DIM].push({
                        sortIndex: sortIndex,
                        direction: item.measureSortItem.direction,
                        type: "data"
                    });
                }
                else {
                    throw new SortDefinitionError("Bad sort - no measureGroup but have measureSortItem");
                }
            }
        });
        return sortActions;
    };
    /**
     * This method evaluates measure sort item locators.
     *
     * @param sortItem measure sort item to evaluate
     * @param dimension dimension in which to evaluate (ROW/COL)
     * @returns index within the dimension to sort by
     */
    FunSort.prototype.findSortIndex = function (sortItem, dimension) {
        var dim = this.result.dims[dimension];
        var headers = this.result.headerItems[dimension];
        var measureHeaders = this.result.measureHeaderItems[dimension];
        var locators = sortItem.measureSortItem.locators;
        // find measure item locator within locators
        var measureLocatorIdx = locators.findIndex(function (locator) { return isMeasureLocatorItem(locator); });
        // it must be there, otherwise this is not measure sort item :/
        if (measureLocatorIdx === -1) {
            throw new SortDefinitionError("Bad sort - no measure locator");
        }
        // single out measure locator definition...
        var measureLocator = locators[measureLocatorIdx];
        // ... and then all attribute locators
        var attributeLocators = locators.filter(function (_, idx) { return idx !== measureLocatorIdx; });
        // go through measure headers and identify all indexes where measure with the desired locator exists
        var possibleMatches = measureHeaders
            .map(function (measure, idx) {
            if (measure.measureHeaderItem.localIdentifier ===
                measureLocator.measureLocatorItem.measureIdentifier) {
                return idx;
            }
            else {
                return -1;
            }
        })
            .filter(function (idx) { return idx > -1; });
        if (possibleMatches.length === 0) {
            throw new SortDefinitionError("Bad sort - measure locator \"" + measureLocator.measureLocatorItem.measureIdentifier + "\" is invalid");
        }
        // find indexes of header groups for attributes used in locators
        var headerGroupIndexes = attributeLocators.map(function (locator) {
            return findHeaderGroupIndexForLocator(locator, dim);
        });
        // evaluate all possible matches...
        var exactMatches = possibleMatches
            .map(function (headerIndex) {
            if (attributeLocators.length === 0) {
                return { match: true, headerIndex: headerIndex };
            }
            // ... see whether all remaining locators match
            var match = attributeLocators.every(function (attributeLocator, locIdx) {
                // find header group index (== index to array with all headers for the attribute) for this locator
                var headerGroupIndex = headerGroupIndexes[locIdx];
                if (headerGroupIndex === -1) {
                    var id = attributeLocator.attributeLocatorItem.attributeIdentifier;
                    throw new SortDefinitionError("Bad sort - attribute \"" + id + "\" not in same dimension as measure");
                }
                // check the item on the column
                var headerItem = headers[headerGroupIndex][headerIndex];
                return (isAttributeHeaderItem(headerItem) &&
                    headerItem.attributeHeaderItem.uri === attributeLocator.attributeLocatorItem.element);
            });
            return { match: match, headerIndex: headerIndex };
        })
            .filter(function (result) { return result.match; });
        if (exactMatches.length > 1) {
            throw new SortDefinitionError("Bad sort - ambiguous locator");
        }
        else if (exactMatches.length === 0) {
            throw new SortDefinitionError("Bad sort - locator does not match a thing");
        }
        return exactMatches[0].headerIndex;
    };
    FunSort.prototype.sortVector = function (sortActions, dimension) {
        var _this = this;
        var dimActions = sortActions[dimension];
        if (dimActions.length === 0) {
            return [];
        }
        var data = this.result.data;
        var entries = lodash_1.zip.apply(void 0, dimActions.map(function (action) {
            if (action.type === "headers") {
                /*
                 * Sorting by headers => values of the desired headers from the respective dimension
                 */
                return _this.result.headerItems[dimension][action.sortIndex].map(function (hi) { return hi.attributeHeaderItem.name; });
            }
            else {
                if (dimension === ROW_DIM) {
                    /*
                     * when sorting rows between each other, extract value from particular column in sheet of data
                     */
                    if (Array.isArray(data[0])) {
                        // two dim data => extract column from each row
                        return data.map(function (row) { return row[action.sortIndex]; });
                    }
                    // one dim data => get value at index
                    return data[action.sortIndex];
                }
                else {
                    if (Array.isArray(data[0])) {
                        // two dim data => extract entire row
                        return data[action.sortIndex];
                    }
                    // one dim data => get the whole row
                    return data;
                }
            }
        })).map(function (values, originalIndex) {
            return {
                values: values,
                originalIndex: originalIndex
            };
        });
        // const sortedEntries = mergeSort(entries, createComparator(dimActions));
        entries.sort(createComparator(dimActions));
        return entries
            .map(function (entry, idx) {
            return {
                from: entry.originalIndex,
                to: idx
            };
        })
            .filter(function (vector) { return vector.from !== vector.to; });
    };
    FunSort.prototype.sort = function (sortDims) {
        var _this = this;
        sortDims.forEach(function (sort, dim) {
            if (sort.length === 0) {
                return;
            }
            var headers = _this.result.headerItems[dim];
            var data = _this.result.data;
            headers.forEach(function (header) {
                _this.sortArray(sort, header);
            });
            _this.sortArray(sort, _this.result.measureHeaderItems[dim]);
            if (dim === ROW_DIM) {
                _this.sortArray(sort, data);
            }
            else {
                data.forEach(function (row) {
                    _this.sortArray(sort, row);
                });
            }
        });
    };
    FunSort.prototype.sortArray = function (sort, array) {
        if (array === undefined || array.length === 0) {
            return;
        }
        var stash = {};
        sort.forEach(function (vector) {
            // stash the element that will be replaced
            stash[vector.to] = array[vector.to];
            var stashed = stash[vector.from];
            // then replace the element
            array[vector.to] = stashed !== undefined ? stashed : array[vector.from];
        });
    };
    return FunSort;
}());
exports.FunSort = FunSort;
function createComparator(sortItems) {
    return function (a, b) {
        var idx = 0;
        for (var _i = 0, sortItems_1 = sortItems; _i < sortItems_1.length; _i++) {
            var item = sortItems_1[_i];
            var left = void 0;
            var right = void 0;
            if (item.type === "data") {
                left = parseFloat(a.values[idx]);
                right = parseFloat(b.values[idx]);
            }
            else {
                left = a.values[idx];
                right = b.values[idx];
            }
            if (left < right) {
                return -1 * (item.direction === "asc" ? 1 : -1);
            }
            else if (left > right) {
                return item.direction === "asc" ? 1 : -1;
            }
            idx++;
        }
        /*
         * Two items are never equal - tie-breakers are solved by comparing placement in the original
         * array. This leads to stable sorting - items with equal sort values are in the same order as
         * they were in the original array.
         */
        return a.originalIndex < b.originalIndex ? -1 : 1;
    };
}
exports.createComparator = createComparator;
function findHeaderGroupIndexForLocator(locator, dim) {
    return findHeaderById(locator.attributeLocatorItem.attributeIdentifier, dim);
}
function findHeaderGroupIndexForItem(sortItem, dim) {
    return findHeaderById(sortItem.attributeSortItem.attributeIdentifier, dim);
}
function findHeaderById(localIdentifier, dim) {
    return dim.headers.findIndex(function (header) {
        return isAttributeHeader(header) && header.attributeHeader.localIdentifier === localIdentifier;
    });
}
//# sourceMappingURL=sorting.js.map