"use strict";
exports.__esModule = true;
exports.exportReport = exports.hasCachedResult = exports.cleanOldCache = void 0;
var lodash_1 = require("lodash");
var HttpStatusCodes = require("http-status-codes");
var executionCache = new Map();
function cleanOldCache(lifetime) {
    if (lifetime === void 0) { lifetime = 240000; }
    var now = Date.now();
    var oldEntries = [];
    executionCache.forEach(function (entry, key) {
        if (entry.createdAt < now - lifetime) {
            oldEntries.push(key);
        }
    });
    oldEntries.forEach(function (entry) { return executionCache["delete"](entry); });
}
exports.cleanOldCache = cleanOldCache;
function hasCachedResult(key) {
    return executionCache.has(key);
}
exports.hasCachedResult = hasCachedResult;
function getPathToExportedFile(format) {
    var path = require("path");
    return path.join(__dirname, "../../static/dummy-export." + format);
}
function getExportResponse(executionId, projectId) {
    return {
        uri: "/gdc/exporter/result/" + projectId + "/" + executionId
    };
}
function getExportContent(format) {
    var fs = require("fs");
    try {
        return fs.readFileSync(getPathToExportedFile(format), "utf8");
    }
    catch (err) {
        return "INTERNAL_SERVER_ERROR";
    }
}
function getExportError(message) {
    return {
        error: {
            trace: "",
            parameters: [""],
            requestId: "",
            component: "",
            errorClass: "GDC::Exception::",
            message: message
        }
    };
}
function createExportResponse(projectId, body, res, executionId) {
    var _a = body.resultExport, executionResult = _a.executionResult, exportConfig = _a.exportConfig;
    if (!executionResult) {
        return res.status(HttpStatusCodes.BAD_REQUEST).json(getExportError("BAD_REQUEST"));
    }
    cleanOldCache();
    executionCache.set(executionId, {
        createdAt: Date.now(),
        exportConfig: exportConfig || {},
        pollCounter: 0
    });
    return res.status(HttpStatusCodes.CREATED).send(getExportResponse(executionId, projectId));
}
function maybeWaitAndCreateResult(projectId, req, res, globalPollCount) {
    var executionId = req.params.id;
    var _a = executionCache.get(executionId), exportConfig = _a.exportConfig, pollCounter = _a.pollCounter;
    if (pollCounter < globalPollCount) {
        executionCache.set(executionId, {
            createdAt: Date.now(),
            exportConfig: exportConfig,
            pollCounter: pollCounter + 1
        });
        return res.status(HttpStatusCodes.ACCEPTED).send(getExportResponse(executionId, projectId));
    }
    var format = exportConfig.format || "csv";
    var title = exportConfig.title || "Untitled";
    var contentDisposition = "attachment; filename=\"" + title + "." + format + "\"; filename*=UTF-8''" + title + "." + format;
    var contentType = format === "csv"
        ? "text/csv;charset=utf8"
        : "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
    return res
        .status(HttpStatusCodes.OK)
        .set("Content-Disposition", contentDisposition)
        .set("Content-Type", contentType)
        .send(getExportContent(format));
}
function downloadFile(req, res) {
    var executionId = req.params.id;
    var exportConfig = executionCache.get(executionId).exportConfig;
    var format = exportConfig.format || "csv";
    var title = exportConfig.title || "Untitled";
    return res.download(getPathToExportedFile(format), title + "." + format);
}
exports.exportReport = {
    register: function (app, project, config, idGenerator) {
        if (idGenerator === void 0) { idGenerator = lodash_1.uniqueId; }
        var projectId = project.project.meta.identifier;
        app.post("/gdc/internal/projects/" + projectId + "/exportResult", function (req, res) {
            return createExportResponse(projectId, req.body, res, idGenerator());
        });
        app.head("/gdc/exporter/result/" + projectId + "/:id", function (req, res) {
            return maybeWaitAndCreateResult(projectId, req, res, config.pollCount);
        });
        app.get("/gdc/exporter/result/" + projectId + "/:id", function (req, res) { return downloadFile(req, res); });
        return app;
    }
};
//# sourceMappingURL=exportReport.js.map