"use strict";
exports.__esModule = true;
exports.createElementsMap = exports.generateDateAttributeElementValues = void 0;
// (C) 2007-2021 GoodData Corporation
var lodash_1 = require("lodash");
var date_fns_1 = require("date-fns");
var objects_1 = require("../helpers/objects");
var routes_1 = require("../../route/routes");
var Project_1 = require("../../model/Project");
var AttributeBuilder_1 = require("./AttributeBuilder");
var ProjectBuilder_1 = require("./ProjectBuilder");
var map_1 = require("../../utils/map");
var DateDataSetBuilder_1 = require("./DateDataSetBuilder");
var DateRange_1 = require("../../utils/model/DateRange");
function createAttributeElements(schema) {
    var project = ProjectBuilder_1.createProject(schema);
    var groups = objects_1.ensureItems(schema.groups);
    return lodash_1.flatMap(groups, function (group) { return objects_1.ensureItems(group.attributes); }).reduce(function (elementMap, attribute) {
        var attributeId = objects_1.ensureIdentifier(attribute, AttributeBuilder_1.createAttributeIdentifier(attribute));
        var elements = objects_1.ensureItems(attribute.elements).map(function (title, i) { return ({
            element: {
                title: title,
                uri: routes_1.elementUri(Project_1.getProjectId(project), attributeId, "" + (i + 1))
            }
        }); });
        return elementMap.set(attributeId, elements);
    }, new Map());
}
function getYearElements(dateRange) {
    var years = Array.from(dateRange.getDatesBetween()).map(function (date) { return date.getFullYear().toString(); });
    return lodash_1.uniq(years);
}
function getMonthElements(dateRange) {
    var months = Array.from(dateRange.getDatesBetween()).map(function (date) { return date_fns_1.format(date, "MMM YYYY"); });
    return lodash_1.uniq(months);
}
function getQuarterElements(dateRange) {
    var quarters = Array.from(dateRange.getDatesBetween()).map(function (date) { return date_fns_1.format(date, "[Q]Q/YYYY"); });
    return lodash_1.uniq(quarters);
}
function getDateElements(dateRange) {
    var dates = Array.from(dateRange.getDatesBetween()).map(function (date) { return date_fns_1.format(date, "MM-DD-YYYY"); });
    return lodash_1.uniq(dates);
}
function getWeekUsElements(dateRange) {
    var weeks = Array.from(dateRange.getDatesBetween()).map(function (date) { return date_fns_1.format(date, "[W]W/YYYY"); });
    return lodash_1.uniq(weeks);
}
function createGenericDateValues(prefix, numberOfValues) {
    return Array.from({ length: numberOfValues }, function (_x, index) { return "" + prefix + (index + 1); });
}
function getMonthInQuarterElements() {
    return createGenericDateValues("M", 3);
}
function getMonthInYearElements() {
    return ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
}
function getQuarterInYearElements() {
    return createGenericDateValues("Q", 4);
}
function getWeekUsInQuarterElements() {
    return createGenericDateValues("W", 14);
}
function getWeekUsInYearElements() {
    return createGenericDateValues("W", 53);
}
function getDayInWeekElements() {
    return ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
}
function getDayInMonthElements() {
    return createGenericDateValues("D", 31);
}
function getDayInQuarterElements() {
    return createGenericDateValues("D", 92);
}
function getDayInYearElements() {
    return createGenericDateValues("D", 366);
}
function generateDateAttributeElementValues(type, dateRange) {
    if (type === "GDC.time.year") {
        return getYearElements(dateRange);
    }
    if (type === "GDC.time.month") {
        return getMonthElements(dateRange);
    }
    if (type === "GDC.time.month_in_quarter") {
        return getMonthInQuarterElements();
    }
    if (type === "GDC.time.month_in_year") {
        return getMonthInYearElements();
    }
    if (type === "GDC.time.quarter") {
        return getQuarterElements(dateRange);
    }
    if (type === "GDC.time.quarter_in_year") {
        return getQuarterInYearElements();
    }
    if (type === "GDC.time.week_us") {
        return getWeekUsElements(dateRange);
    }
    if (type === "GDC.time.week_in_quarter") {
        return getWeekUsInQuarterElements();
    }
    if (type === "GDC.time.week_in_year") {
        return getWeekUsInYearElements();
    }
    if (type === "GDC.time.date") {
        return getDateElements(dateRange);
    }
    if (type === "GDC.time.date") {
        return getDateElements(dateRange);
    }
    if (type === "GDC.time.day_in_week") {
        return getDayInWeekElements();
    }
    if (type === "GDC.time.day_in_month") {
        return getDayInMonthElements();
    }
    if (type === "GDC.time.day_in_quarter") {
        return getDayInQuarterElements();
    }
    if (type === "GDC.time.day_in_year") {
        return getDayInYearElements();
    }
    return [];
}
exports.generateDateAttributeElementValues = generateDateAttributeElementValues;
function createDateAttributeElements(schema) {
    var project = ProjectBuilder_1.createProject(schema);
    var groups = objects_1.ensureItems(schema.groups);
    var dateDataSets = lodash_1.flatMap(groups, function (group) { return objects_1.ensureItems(group.dateDataSets); });
    var map = new Map();
    for (var _i = 0, dateDataSets_1 = dateDataSets; _i < dateDataSets_1.length; _i++) {
        var dateDataSetSchema = dateDataSets_1[_i];
        var dateDataSet = DateDataSetBuilder_1.createDateDataSet(project, dateDataSetSchema);
        var startDate = dateDataSetSchema.startDate, finalDate = dateDataSetSchema.finalDate;
        var _loop_1 = function (fragment) {
            var elementValues = generateDateAttributeElementValues(fragment.type, new DateRange_1.DateRange(startDate, finalDate));
            var elements = elementValues.map(function (value, i) { return ({
                element: {
                    title: value,
                    uri: routes_1.elementUri(Project_1.getProjectId(project), fragment.meta.identifier, "" + (i + 1))
                }
            }); });
            map.set(fragment.meta.identifier, elements);
        };
        for (var _a = 0, _b = dateDataSet.availableDateAttributes; _a < _b.length; _a++) {
            var fragment = _b[_a];
            _loop_1(fragment);
        }
    }
    return map;
}
function createElementsMap(schema) {
    var attributeMap = createAttributeElements(schema);
    var dateMap = createDateAttributeElements(schema);
    return map_1.merge(attributeMap, dateMap);
}
exports.createElementsMap = createElementsMap;
//# sourceMappingURL=ElementsBuilder.js.map