// (C) 2019-2020 GoodData Corporation
import { AFM } from "@gooddata/typings";
import { appendFilters, dateFiltersDataSetsMatch, getId, getDateFilterDateDataSet } from "../mergeFilters";
import * as fixture from "./fixtures/Afm.fixtures";

describe("AFM utils", () => {
    describe("appendFilters", () => {
        it("should concatenate filters when all different", () => {
            const afm = {
                filters: [fixture.positiveAttributeFilter],
            };
            const attributeFilters = [fixture.negativeAttributeFilter];
            const enriched = appendFilters(afm, attributeFilters, fixture.relativeDateFilter);
            expect(enriched.filters).toEqual([
                fixture.positiveAttributeFilter,
                fixture.negativeAttributeFilter,
                fixture.relativeDateFilter,
            ]);
        });

        it("should override date filter if identifier is identical", () => {
            const afm = {
                filters: [fixture.positiveAttributeFilter, fixture.relativeDateFilter],
            };
            const attributeFilters: AFM.AttributeFilterItem[] = [];
            const enriched = appendFilters(afm, attributeFilters, fixture.relativeDateFilter);
            expect(enriched.filters).toEqual([fixture.positiveAttributeFilter, fixture.relativeDateFilter]);
        });

        it("should append date filter if ID is different", () => {
            const afm = {
                filters: [fixture.relativeDateFilter],
            };
            const attributeFilters: AFM.AttributeFilterItem[] = [];
            const enriched = appendFilters(afm, attributeFilters, fixture.absoluteDateFilter2);
            expect(enriched.filters).toEqual([fixture.absoluteDateFilter2, fixture.relativeDateFilter]);
        });

        it('should delete date filter from AFM if "All time" date filter requested', () => {
            const afm = {
                filters: [fixture.relativeDateFilter],
            };
            const enriched = appendFilters(afm, [], fixture.allTimeDateFilter);
            expect(enriched.filters).toEqual([]);
        });

        it("should add date filter to empty afm filters", () => {
            const afm: AFM.IAfm = {
                filters: [],
            };
            const enriched = appendFilters(afm, [], fixture.absoluteDateFilter1);
            expect(enriched.filters).toEqual([fixture.absoluteDateFilter1]);
        });

        it("should work without filters in afm", () => {
            const afm: AFM.IAfm = {};
            const enriched = appendFilters(afm, [], fixture.absoluteDateFilter1);
            expect(enriched.filters).toEqual([fixture.absoluteDateFilter1]);
        });

        it("should not modify afm if there are no filters", () => {
            const afm: AFM.IAfm = {};
            const expectedAfm: AFM.IAfm = {};
            const enriched = appendFilters(afm, [], undefined);
            expect(enriched).toEqual(expectedAfm);
        });

        it("should return measure value filters if only these filters are provided", () => {
            const afm: AFM.IAfm = {};
            const enriched = appendFilters(afm, [], undefined, [fixture.measureValueFilter1]);
            expect(enriched.filters).toEqual([fixture.measureValueFilter1]);
        });

        it("should return ranking filters if only these filters are provided", () => {
            const afm: AFM.IAfm = {};
            const enriched = appendFilters(afm, [], undefined, undefined, [fixture.rankingFilter1]);
            expect(enriched.filters).toEqual([fixture.rankingFilter1]);
        });

        it("should return all filters merged when all types are provided", () => {
            const afm = {
                filters: [
                    fixture.absoluteDateFilter1,
                    fixture.positiveAttributeFilter,
                    fixture.measureValueFilter1,
                    fixture.rankingFilter1,
                ],
            };
            const enriched = appendFilters(
                afm,
                [fixture.negativeAttributeFilter],
                fixture.absoluteDateFilter2,
                [fixture.measureValueFilter2],
                [fixture.rankingFilter2],
            );
            expect(enriched.filters).toEqual([
                fixture.positiveAttributeFilter,
                fixture.measureValueFilter1,
                fixture.rankingFilter1,
                fixture.negativeAttributeFilter,
                fixture.absoluteDateFilter2,
                fixture.absoluteDateFilter1,
                fixture.measureValueFilter2,
                fixture.rankingFilter2,
            ]);
        });
    });

    describe("dateFiltersDataSetsMatch", () => {
        it("should match same filters", () => {
            expect(dateFiltersDataSetsMatch(fixture.relativeDateFilter, fixture.allTimeDateFilter)).toEqual(
                true,
            );

            expect(dateFiltersDataSetsMatch(fixture.relativeDateFilter, fixture.absoluteDateFilter1)).toEqual(
                true,
            );
        });

        it("should detect different filters", () => {
            expect(
                dateFiltersDataSetsMatch(fixture.absoluteDateFilter1, fixture.absoluteDateFilter2),
            ).toEqual(false);
        });
    });

    describe("getId", () => {
        it("should return uri if uri object qualifier is provided", () => {
            const result = getId({
                uri: "/uri",
            });
            expect(result).toEqual("/uri");
        });

        it("should return identifier if identifier object qualifier is provided", () => {
            const result = getId({
                identifier: "id",
            });
            expect(result).toEqual("id");
        });
    });

    describe("getDateFilterDateDataSet", () => {
        it("should return data set of relative date filter", () => {
            const filter = fixture.relativeDateFilter;

            const dataSet = getDateFilterDateDataSet(filter);

            expect(dataSet).toBe(filter.relativeDateFilter.dataSet);
        });

        it("should return data set of absolute date filter", () => {
            const filter = fixture.absoluteDateFilter1;

            const dataSet = getDateFilterDateDataSet(filter);

            expect(dataSet).toBe(filter.absoluteDateFilter.dataSet);
        });

        it("should throw exception for invalid date filter", () => {
            const filter = {
                anotherDateFilter: {
                    dataSet: {
                        uri: "/gdc/md/project/obj/727",
                    },
                },
            };

            // noinspection JSValidateTypesInspection
            const getDataSet = () => getDateFilterDateDataSet(filter as any);

            expect(getDataSet).toThrow();
        });
    });
});
